﻿//---------------------------------------------------------------------
// This file is part of ediFabric
//
// Copyright (c) ediFabric. All rights reserved.
//
// THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
// KIND, WHETHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
// PURPOSE.
//---------------------------------------------------------------------

namespace EdiFabric.Framework.Helpers
{
    using System;
    using System.Collections.Generic;
    using System.Collections;
    using System.IO;
    using System.Linq;
    using System.Xml.Linq;
    using System.Xml.Serialization;

    using EdiFabric.Framework.Constants;
    using EdiFabric.Framework.Entities;

    class EdiHelper
    {
        public static MemoryStream CopyStream(Stream ediStream)
        {
            if (ediStream == null) throw new ArgumentNullException("ediStream");

            var tempStream = new MemoryStream();
            ediStream.CopyTo(tempStream);
            ediStream.Position = tempStream.Position = 0;

            return tempStream;
        }

        public static XElement Serialize<T>(T instance)
        {
            var serializer = new XmlSerializer(typeof(T), typeof(T).Namespace);
            using (var ms = new MemoryStream())
            {
                serializer.Serialize(ms, instance);
                ms.Position = 0;
                return XElement.Load(ms);
            }
        }

        public static T Deserialize<T>(XElement xml)
        {
            var serializer = new XmlSerializer(typeof(T), typeof(T).Namespace);
            return (T)serializer.Deserialize(xml.CreateReader());
        }

        public static string[] GetEdiSegments(Stream ediStream, InterchangeContext interchangeContext)
        {
            if (ediStream == null) throw new ArgumentNullException("ediStream");
            if (interchangeContext == null) throw new ArgumentNullException("interchangeContext");

            using (var reader = new StreamReader(ediStream, false))
            {
                var contents = reader.ReadToEnd();
                if (interchangeContext.SegmentTerminator != Environment.NewLine) contents = contents.Replace(Environment.NewLine, string.Empty);
                var result = contents.Split(interchangeContext.SegmentTerminator.ToCharArray(), StringSplitOptions.RemoveEmptyEntries);
                result = (from r in result select r.Trim()).ToArray();
                return result;
            }
        }

        public static string[] GetEdiCompositeDataElements(string ediSegment, InterchangeContext interchangeContext)
        {
            if (string.IsNullOrEmpty(ediSegment)) throw new ArgumentNullException("ediSegment");
            if (interchangeContext == null) throw new ArgumentNullException("interchangeContext");

            var result = ediSegment.Split(interchangeContext.DataElementSeparator.ToCharArray(), StringSplitOptions.None);
            if (interchangeContext.Format != Format.X12)
            {
                result = ManageEscapeCharacters(result, interchangeContext.ReleaseIndicator, interchangeContext.DataElementSeparator);
            }
            //  Skip segment name
            return result.Skip(1).ToArray();
        }

        public static string[] GetEdiComponentDataElements(string ediCompositeDataElement, InterchangeContext interchangeContext)
        {
            if (interchangeContext == null) throw new ArgumentNullException("interchangeContext");
            if (string.IsNullOrEmpty(ediCompositeDataElement)) throw new ArgumentNullException("ediCompositeDataElement");

            var result = ediCompositeDataElement.Split(interchangeContext.ComponentDataElementSeparator.ToCharArray(), StringSplitOptions.None);
            if (interchangeContext.Format != Format.X12)
            {
                result = ManageEscapeCharacters(result, interchangeContext.ReleaseIndicator, interchangeContext.ComponentDataElementSeparator);
            }

            return result;
        }

        public static string GetSegmentName(string ediSegment, InterchangeContext interchangeContext)
        {
            if (string.IsNullOrEmpty(ediSegment)) throw new ArgumentNullException("ediSegment");
            if (interchangeContext == null) throw new ArgumentNullException("interchangeContext");
            if (ediSegment.StartsWith(Segment.Una)) return Segment.Una;

            return ediSegment.Split(interchangeContext.DataElementSeparator.ToCharArray(), StringSplitOptions.None)[0];
        }

        public static string GetElementName(XElement element, InterchangeContext interchangeContext)
        {
            if (element == null) throw new ArgumentNullException("element");
            if (interchangeContext == null) throw new ArgumentNullException("interchangeContext");

            return element.Name.LocalName.Split(interchangeContext.DataElementSeparator.ToCharArray(), StringSplitOptions.None)[0].Split(new[] { '_' })[1];
        }

        private static string[] ManageEscapeCharacters(IEnumerable<string> stringArray, string escapeCharacter, string splitSeparator)
        {
            if (stringArray == null) throw new ArgumentNullException("stringArray");
            if (string.IsNullOrEmpty(escapeCharacter)) throw new ArgumentNullException("escapeCharacter");
            if (string.IsNullOrEmpty(splitSeparator)) throw new ArgumentNullException("splitSeparator");

            var currentList = new List<string>();

            string toBeEscaped = null;

            IEnumerator enumerator = stringArray.GetEnumerator();
            while (enumerator.MoveNext())
            {
                var currentString = (string)enumerator.Current;

                if (currentString.EndsWith(escapeCharacter))
                {
                    toBeEscaped = currentString;
                }
                else
                {
                    if (toBeEscaped != null)
                    {
                        toBeEscaped = toBeEscaped.Replace(escapeCharacter, splitSeparator);
                        currentString = toBeEscaped + currentString;
                    }

                    currentList.Add(currentString);
                    toBeEscaped = null;
                }
            }

            var result = new string[currentList.Count];

            currentList.CopyTo(result);

            return result;
        }
    }
}
