﻿//---------------------------------------------------------------------
// This file is part of ediFabric
//
// Copyright (c) ediFabric. All rights reserved.
//
// THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
// KIND, WHETHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
// PURPOSE.
//---------------------------------------------------------------------

namespace EdiFabric.Framework.Helpers
{
    using System;
    using System.Globalization;
    using System.Linq;
    using System.IO;
    using System.Xml.Linq;

    using EdiFabric.Framework.Constants;
    using EdiFabric.Framework.Entities;
    using EdiFabric.Framework.Exceptions;

    class InterchangeContextHelper
    {
        public static InterchangeContext GetInterchangeContext(MemoryStream ediStream)
        {
            if (ediStream == null) throw new ArgumentNullException("ediStream");

            var result = new InterchangeContext();
            using (var reader = new StreamReader(ediStream, false))
            {
                var contents = reader.ReadToEnd();
                contents = contents.Replace(Environment.NewLine, string.Empty);

                if (contents.ToUpper().StartsWith(Segment.Isa))
                {
                    try
                    {
                        result.DataElementSeparator = contents[3].ToString(CultureInfo.InvariantCulture);
                        var isa = string.Concat(contents.Take(106));
                        var isaElements = isa.Split(result.DataElementSeparator[0]);
                        result.ComponentDataElementSeparator = string.Concat(isaElements[16].First());
                        result.RepetitionSeparator = isaElements[11] != "U" ? isaElements[11] : "^";
                        var st = isaElements[16].Skip(1).DefaultIfEmpty('G').First();
                        result.SegmentTerminator = st.ToString(CultureInfo.InvariantCulture);
                        if (result.SegmentTerminator == " " || string.IsNullOrEmpty(result.SegmentTerminator)
                            || result.SegmentTerminator == "G") result.SegmentTerminator = Environment.NewLine;
                        result.ReleaseIndicator = "";
                        result.Format = Format.X12;
                        result.TargetNamespace = Namespace.TargetNamespaceX12;

                        return result;
                    }
                    catch
                    {
                        throw new ParserException("Can't find X12 interchange delimiters");
                    }
                }

                if (contents.ToUpper().StartsWith(Segment.Unb))
                {
                    //  Default Edifact
                    result.ComponentDataElementSeparator = ":";
                    result.DataElementSeparator = "+";
                    result.ReleaseIndicator = "?";
                    result.RepetitionSeparator = "*";
                    result.SegmentTerminator = "'";
                    result.Format = Format.Edifact;
                    result.TargetNamespace = Namespace.TargetNamespaceEdifact;

                    return result;
                }

                if (contents.ToUpper().StartsWith(Segment.Una))
                {
                    try
                    {
                        //  Parse UNA terminators
                        var una = contents.Replace(Segment.Una, "").Take(6).ToList();

                        result.ComponentDataElementSeparator = una[0].ToString(CultureInfo.InvariantCulture);
                        result.DataElementSeparator = una[1].ToString(CultureInfo.InvariantCulture);
                        result.ReleaseIndicator = una[3].ToString(CultureInfo.InvariantCulture);
                        result.RepetitionSeparator = "*";
                        result.SegmentTerminator = una[5].ToString(CultureInfo.InvariantCulture);
                        result.Format = Format.Edifact;
                        result.TargetNamespace = Namespace.TargetNamespaceEdifact;

                        return result;
                    }
                    catch
                    {
                        throw new ParserException("Can't find UNA interchange delimiters");
                    }
                }

                throw new ParserException(
                    "Can't recognize EDI format: No " + Segment.Isa + " or " + Segment.Una
                    + " or " + Segment.Unb + " segment has been found at the top line.");
            }
        }

        public static InterchangeContext GetInterchangeContext(string envelopeNamespace)
        {
            if (string.IsNullOrEmpty(envelopeNamespace)) throw new ArgumentNullException("envelopeNamespace");

            var result = new InterchangeContext();
            switch (envelopeNamespace)
            {
                case Namespace.TargetNamespaceEdifact:
                    result.ComponentDataElementSeparator = ":";
                    result.DataElementSeparator = "+";
                    result.ReleaseIndicator = "?";
                    result.RepetitionSeparator = "*";
                    result.SegmentTerminator = "'";
                    result.Format = Format.Edifact;
                    break;
                case Namespace.TargetNamespaceX12:
                    result.ComponentDataElementSeparator = ":";
                    result.DataElementSeparator = "*";
                    result.ReleaseIndicator = "";
                    result.RepetitionSeparator = "^";
                    result.SegmentTerminator = "~";
                    result.Format = Format.X12;
                    break;
                default:
                    throw new ParserException("Unsupported format: " + envelopeNamespace);
            }

            return result;
        }

        public static InterchangeContext MergeInterchangeContext(XElement element, InterchangeContext interchangeContext)
        {
            if (element == null) throw new ArgumentNullException("element");
            
            var result = GetInterchangeContext(element.Name.NamespaceName);
            if (interchangeContext == null) return result;

            if (!string.IsNullOrEmpty(interchangeContext.ComponentDataElementSeparator)) result.ComponentDataElementSeparator = interchangeContext.ComponentDataElementSeparator;
            if (!string.IsNullOrEmpty(interchangeContext.DataElementSeparator)) result.DataElementSeparator = interchangeContext.DataElementSeparator;
            if (!string.IsNullOrEmpty(interchangeContext.ReleaseIndicator)) result.ReleaseIndicator = interchangeContext.ReleaseIndicator;
            if (!string.IsNullOrEmpty(interchangeContext.RepetitionSeparator)) result.RepetitionSeparator = interchangeContext.RepetitionSeparator;
            if (!string.IsNullOrEmpty(interchangeContext.SegmentTerminator)) result.SegmentTerminator = interchangeContext.SegmentTerminator;

            return result;
        }
    }
}
