﻿//---------------------------------------------------------------------
// This file is part of ediFabric
//
// Copyright (c) ediFabric. All rights reserved.
//
// THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
// KIND, WHETHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
// PURPOSE.
//---------------------------------------------------------------------

namespace EdiFabric.Framework.Helpers
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Linq;
    using System.Security;
    using System.Xml.Linq;

    using EdiFabric.Framework.Constants;
    using EdiFabric.Framework.Entities;
    using EdiFabric.Framework.Exceptions;

    class ParserHelper
    {
        public static ParseTree GetParsingGrammarForMessage(MessageContext messageContext)
        {
            if (messageContext == null) throw new ArgumentNullException("messageContext");
            return GetParseTree(
                messageContext.TypeName + ".M_" + messageContext.Tag,
                messageContext.AssemblyName,
                pt => pt.IsMessage || pt.IsGroup);
        }

        public static ParseTree GetParsingGrammarForSegment(ParseTree element)
        {
            if (element == null) throw new ArgumentNullException("element");
            return GetParseTree(element.TypeName, element.AssemblyName, pt => pt.IsSegment || pt.IsComplex);
        }

        public static ParseTree GetParsingGrammarForEnvelopeSegment<T>(InterchangeContext interchangeContext)
        {
            return
                GetParsingGrammarForSegment(
                    new ParseTree { TypeName = Namespace.EnvelopeNamespace + interchangeContext.Format + "." + typeof(T).Name });
        }

        public static XElement ParseSegment(ParsingContext segment, InterchangeContext interchangeContext)
        {
            if (segment == null) throw new ArgumentNullException("segment");
            if (interchangeContext == null) throw new ArgumentNullException("interchangeContext");
            if (!segment.ContextEntity.IsSegment) throw new Exception("Not a segment.");

            XNamespace ns = interchangeContext.TargetNamespace;
            var result = new XElement(ns + segment.ContextEntity.Name);
            var dataElements = EdiHelper.GetEdiCompositeDataElements(segment.ContextLine, interchangeContext).ToList();
            var indexed = dataElements.Select((a, i) => new { Item = a, Position = i }).ToList();

            foreach (var dataElement in indexed)
            {
                if (string.IsNullOrEmpty(dataElement.Item)) continue;

                var elementGrammar = dataElement.Position >= segment.ContextEntity.Children.Count
                                         ? segment.ContextEntity.Children.Last()
                                         : segment.ContextEntity.Children[dataElement.Position];

                var elementRepetitions = dataElement.Item.Split(interchangeContext.RepetitionSeparator.ToCharArray());

                foreach (var elementRepetition in elementRepetitions)
                {
                    result.Add(ParseElement(elementGrammar, elementRepetition, interchangeContext));
                }
            }

            return result;
        }

        public static List<string> ParseSegment(XElement segment, ParseTree parseTree, InterchangeContext interchangeContext)
        {
            var result = new List<string>();

            if (segment.Name.LocalName.StartsWith("G_"))
            {
                foreach (var childSegment in segment.Elements())
                {
                    var segmentGrammar =
                        GetParsingGrammarForSegment(
                            new ParseTree
                                {
                                    TypeName = parseTree.TypeName.Replace(parseTree.Name, childSegment.Name.LocalName),
                                    AssemblyName = parseTree.AssemblyName
                                });
                    result.AddRange(ParseSegment(childSegment, segmentGrammar, interchangeContext));
                }

                return result;
            }

            if (segment.Name.LocalName.StartsWith("S_"))
            {
                result.Add(ParseDataElements(segment, parseTree, interchangeContext));
                return result;
            }

            throw new ParserException("Can't parse segment " + segment.Name);
        }

        public static T PopulateSegment<T>(string segment, InterchangeContext interchangeContext)
        {
            if (segment == null) throw new ArgumentNullException("segment");

            var grammar = GetParsingGrammarForSegment(new ParseTree { TypeName = Namespace.EnvelopeNamespace + interchangeContext.Format + "." + typeof(T).Name });
            var values = EdiHelper.GetEdiCompositeDataElements(segment, interchangeContext).ToList();

            XNamespace ns = interchangeContext.TargetNamespace;
            var result = new XElement(ns + grammar.Name);

            var indexedGrammar = grammar.Children.Select((g, p) => new { Grammar = g, Position = p }).ToList();
            var indexedValues = values.Select((v, p) => new { Value = v, Position = p }).ToList();

            var indexedList = from ig in indexedGrammar 
                    from iv in indexedValues 
                    where ig.Position == iv.Position
                    select new { ig.Grammar, iv.Value };

            foreach (var element in indexedList)
            {
                result.Add(PopulateElement(element.Grammar, element.Value, interchangeContext));
            }

            return EdiHelper.Deserialize<T>(result);
        }

        private static XElement ParseElement(ParseTree elementGrammar, string value, InterchangeContext interchangeContext)
        {
            if (elementGrammar == null) throw new ArgumentNullException("elementGrammar");
            if (interchangeContext == null) throw new ArgumentNullException("interchangeContext");
            if (value == null) throw new ArgumentNullException("value");
            if (!elementGrammar.IsComplex && !elementGrammar.IsSimple)throw new Exception("Not a data element.");

            XNamespace ns = interchangeContext.TargetNamespace;
            var result = new XElement(ns + elementGrammar.Name);

            if (elementGrammar.IsComplex)
            {
                var componentDataElements = EdiHelper.GetEdiComponentDataElements(value, interchangeContext).ToList();
                var indexed = componentDataElements.Select((a, i) => new { Item = a, Position = i }).ToList();

                foreach (var dataElement in indexed)
                {
                    if (string.IsNullOrEmpty(dataElement.Item)) continue;

                    var objectToParse = dataElement.Position >= elementGrammar.Children.Count
                                            ? elementGrammar.Children.Last()
                                            : elementGrammar.Children[dataElement.Position];

                    var element = new XElement(ns + objectToParse.Name);
                    element.SetValue(SecurityElement.Escape(dataElement.Item) ?? string.Empty);
                    result.Add(element);
                }
            }
            else
            {
                result.SetValue(SecurityElement.Escape(value) ?? string.Empty);
            }

            return result;
        }

        private static XElement PopulateElement(ParseTree elementGrammar, string value, InterchangeContext interchangeContext)
        {
            if (elementGrammar == null) throw new ArgumentNullException("elementGrammar");
            if (interchangeContext == null) throw new ArgumentNullException("interchangeContext");
            if (value == null) throw new ArgumentNullException("value");
            if (!elementGrammar.IsComplex && !elementGrammar.IsSimple) throw new Exception("Not a data element.");

            XNamespace ns = interchangeContext.TargetNamespace;
            var result = new XElement(ns + elementGrammar.Name);

            if (elementGrammar.IsComplex)
            {
                if (value == string.Empty)
                {
                    foreach (var dataElement in elementGrammar.Children)
                    {
                        var element = new XElement(ns + dataElement.Name);
                        element.SetValue(string.Empty);
                        result.Add(element);
                    }
                }
                else
                {
                    var componentDataElements = EdiHelper.GetEdiComponentDataElements(value, interchangeContext).ToList();
                    var indexedGrammar = elementGrammar.Children.Select((g, p) => new { Grammar = g, Position = p }).ToList();
                    var indexedValues = componentDataElements.Select((v, p) => new { Value = v, Position = p }).ToList();
                    var indexedList = from ig in indexedGrammar
                                      from iv in indexedValues
                                      where ig.Position == iv.Position
                                      select new { ig.Grammar, iv.Value };

                    foreach (var dataElement in indexedList)
                    {
                        var element = new XElement(ns + dataElement.Grammar.Name);
                        element.SetValue(SecurityElement.Escape(dataElement.Value) ?? string.Empty);
                        result.Add(element);
                    }
                }
            }
            else
            {
                result.SetValue(SecurityElement.Escape(value) ?? string.Empty);
            }

            return result;
        }

        private static ParseTree GetParseTree(string typeName, string assemblyName, Func<ParseTree, bool> limit)
        {
            if (string.IsNullOrEmpty(typeName)) throw new ArgumentNullException("typeName");
            if (limit == null) throw new ArgumentNullException("limit");
            if (assemblyName == null) assemblyName = string.Empty;

            var ediElement = Type.GetType(typeName + assemblyName);
            
            if (ediElement == null)
            {
                throw new ParserException("Can't find defnition for " + typeName);
            }

            var result = new ParseTree
                {
                    Name = ediElement.Name,
                    Children = new List<ParseTree>(),
                    TypeName = typeName,
                    AssemblyName = assemblyName
                };

            if (limit(result))
            {
                var properties = ediElement.GetProperties();
                foreach (var propertyInfo in properties)
                {
                    var childTypeName = propertyInfo.PropertyType.FullName;

                    if (typeof(IList).IsAssignableFrom(propertyInfo.PropertyType)
                        && propertyInfo.PropertyType.IsGenericType)
                    {
                        childTypeName = propertyInfo.PropertyType.GetGenericArguments()[0].FullName;
                    }

                    var child = propertyInfo.PropertyType.FullName == "System.String"
                                    ? new ParseTree { Name = propertyInfo.Name }
                                    : GetParseTree(childTypeName, assemblyName, limit);

                    child.Parent = result;
                    result.Children.Add(child);
                }
            }

            return result;
        }

        private static string ParseDataElements(XElement segment, ParseTree parseTree, InterchangeContext interchangeContext)
        {
            var result = parseTree.EdiName + interchangeContext.DataElementSeparator;

            foreach (var element in parseTree.Children)
            {
                var currentElement = element;
                var matchingElements = segment.Elements().Where(e => e.Name.LocalName == currentElement.Name);

                string line = string.Empty;
                foreach(var matchingElement in matchingElements)
                {
                    line = line + ParseDataElement(matchingElement, element, interchangeContext) + interchangeContext.RepetitionSeparator;
                }

                line = line.TrimEnd(interchangeContext.RepetitionSeparator.ToCharArray());

                result = result + line + interchangeContext.DataElementSeparator;
            }

            result = result.TrimEnd(interchangeContext.DataElementSeparator.ToCharArray());
            result = result + interchangeContext.SegmentTerminator;
            
            return result;
        }

        private static string ParseDataElement(XElement dataElement, ParseTree parseTree, InterchangeContext interchangeContext)
        {
            string result = string.Empty;

            if (parseTree.IsComplex)
            {
                foreach (var element in parseTree.Children)
                {
                    var currentElement = element;
                    var matchingElement = dataElement.Elements().SingleOrDefault(e => e.Name.LocalName == currentElement.Name);
                    var currentValue = matchingElement == null ? string.Empty : matchingElement.Value;

                    if (currentValue.Contains("&"))
                    {
                        var securityElement = SecurityElement.FromString("<parse>" + currentValue + "</parse>");
                        if (securityElement == null)
                        {
                            throw new ParserException("Can't parse value " + currentValue);
                        }

                        currentValue = securityElement.Text;
                    }

                    result = result + currentValue + interchangeContext.ComponentDataElementSeparator;
                }

                result = result.TrimEnd(interchangeContext.ComponentDataElementSeparator.ToCharArray());
            }
            else
            {
                var valueToSet = dataElement.Value;
                if (valueToSet.Contains("&"))
                {
                    var securityElement = SecurityElement.FromString("<parse>" + valueToSet + "</parse>");
                    if (securityElement == null)
                    {
                        throw new ParserException("Can't parse value " + valueToSet);
                    }

                    valueToSet = securityElement.Text;
                }

                result = valueToSet;
            }

            return result;
        }
    }
}
