﻿//---------------------------------------------------------------------
// This file is part of ediFabric
//
// Copyright (c) ediFabric. All rights reserved.
//
// THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
// KIND, WHETHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
// PURPOSE.
//---------------------------------------------------------------------

namespace EdiFabric.Framework.Parsers.Edi
{
    using System;
    using System.Collections.Generic;
    using System.Xml.Linq;

    using EdiFabric.Framework.Constants;
    using EdiFabric.Framework.Entities;
    using EdiFabric.Framework.Exceptions;
    using EdiFabric.Framework.Helpers;
    using EdiFabric.Framework.InterchangeEnvelopes.Edifact;

    class EdifactBuilder : EdiBuilder
    {
        public EdifactBuilder(InterchangeContext interchangeContext)
        {
            if (interchangeContext == null) throw new ArgumentNullException("interchangeContext");

            this.interchangeContext = interchangeContext;
            this.ediSegments = new List<string>();
        }

        public override void CreateInterchangeSettings(XElement xmlEdi)
        {
            var interchangeSettings = InterchangeContextHelper.GetInterchangeContext(xmlEdi.Name.NamespaceName);

            if (!interchangeSettings.Equals(this.interchangeContext))
            {
                var una = Segment.Una + this.interchangeContext.ComponentDataElementSeparator
                          + this.interchangeContext.DataElementSeparator + "."
                          + this.interchangeContext.ReleaseIndicator + " "
                          + this.interchangeContext.SegmentTerminator;
                
                this.ediSegments.Add(una);
            }
        }

        public override void CreateInterchangeHeader(XElement xmlEdi)
        {
            XNamespace ns = xmlEdi.Name.NamespaceName;
            var unb = xmlEdi.Element(ns + "S_" + Segment.Unb);

            if(unb == null)
            {
                throw new ParserException("Can't find UNB segment.");
            }

            this.ediSegments.AddRange(ParserHelper.ParseSegment(unb, ParserHelper.GetParsingGrammarForEnvelopeSegment<S_UNB>(interchangeContext), interchangeContext));
        }

        public override void CreateGroupHeader(XElement group)
        {
            XNamespace ns = group.Name.NamespaceName;
            var ung = group.Element(ns + "S_" + Segment.Ung);

            if(ung != null)
            {
                this.ediSegments.AddRange(ParserHelper.ParseSegment(ung, ParserHelper.GetParsingGrammarForEnvelopeSegment<S_UNG>(interchangeContext), interchangeContext));
            }
        }

        public override void CreateSegment(XElement segment, XElement typedMessage, string origin)
        {
            var messageContext = CreateMessageContext(typedMessage, origin);
            var segmentGrammar =
                ParserHelper.GetParsingGrammarForSegment(
                    new ParseTree
                    {
                        TypeName = messageContext.TypeName + "." + segment.Name.LocalName,
                        AssemblyName = messageContext.AssemblyName
                    });

            this.ediSegments.AddRange(
                ParserHelper.ParseSegment(segment, segmentGrammar, interchangeContext));
        }

        public override void CreateGroupTrailer(XElement group)
        {
            XNamespace ns = group.Name.NamespaceName;
            var une = group.Element(ns + "S_" + Segment.Une);

            if (une != null)
            {
                this.ediSegments.AddRange(ParserHelper.ParseSegment(une, ParserHelper.GetParsingGrammarForEnvelopeSegment<S_UNE>(interchangeContext), interchangeContext));
            }
        }

        public override void CreateInterchangeTrailer(XElement xmlEdi)
        {
            XNamespace ns = xmlEdi.Name.NamespaceName;
            var unz = xmlEdi.Element(ns + "S_" + Segment.Unz);

            if (unz == null)
            {
                throw new ParserException("Can't find UNZ segment.");
            }

            this.ediSegments.AddRange(ParserHelper.ParseSegment(unz, ParserHelper.GetParsingGrammarForEnvelopeSegment<S_UNZ>(interchangeContext), interchangeContext));
        }

        private MessageContext CreateMessageContext(XElement message, string origin)
        {
            var result = new MessageContext();
            
            XNamespace ns = message.Name.NamespaceName;
            var unh = message.Element(ns + "S_" + Segment.Unh);

            if (unh == null)
            {
                throw new ParserException("Can't find UNH segment.");
            }

            var unhContext = unh.Element(ns + Internals.UnhContext);

            if (unhContext == null)
            {
                throw new ParserException("Can't find " + Internals.UnhContext);
            }

            var tag = unhContext.Element(ns + Internals.UnhTag);

            if (tag == null)
            {
                throw new ParserException("Can't find " + Internals.UnhTag);
            }

            var edition = unhContext.Element(ns + Internals.UnhEdition);

            if (edition == null)
            {
                throw new ParserException("Can't find " + Internals.UnhEdition);
            }

            var release = unhContext.Element(ns + Internals.UnhRelease);

            if (release == null)
            {
                throw new ParserException("Can't find " + Internals.UnhRelease);
            }

            result.Tag = tag.Value;
            result.Version = edition.Value + release.Value;
            result.Origin = origin;
            result.Format = Format.Edifact;

            return result;
        }
    }
}
