﻿//---------------------------------------------------------------------
// This file is part of ediFabric
//
// Copyright (c) ediFabric. All rights reserved.
//
// THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
// KIND, WHETHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
// PURPOSE.
//---------------------------------------------------------------------

namespace EdiFabric.Framework.Parsers.Edi
{
    using System;
    using System.Collections.Generic;
    using System.Xml.Linq;

    using EdiFabric.Framework.Constants;
    using EdiFabric.Framework.Entities;
    using EdiFabric.Framework.Exceptions;
    using EdiFabric.Framework.Helpers;
    using EdiFabric.Framework.InterchangeEnvelopes.X12;

    class X12Builder : EdiBuilder
    {
        public X12Builder(InterchangeContext interchangeContext)
        {
            if (interchangeContext == null) throw new ArgumentNullException("interchangeContext");

            this.interchangeContext = interchangeContext;
            this.ediSegments = new List<string>();
        }

        public override void CreateInterchangeSettings(XElement xmlEdi)
        {
        }

        public override void CreateInterchangeHeader(XElement xmlEdi)
        {
            XNamespace ns = xmlEdi.Name.NamespaceName;
            var isa = xmlEdi.Element(ns + "S_" + Segment.Isa);

            if (isa == null)
            {
                throw new ParserException("Can't find ISA segment.");
            }

            this.ediSegments.AddRange(ParserHelper.ParseSegment(isa, ParserHelper.GetParsingGrammarForEnvelopeSegment<S_ISA>(interchangeContext), interchangeContext));
        }

        public override void CreateGroupHeader(XElement group)
        {
            XNamespace ns = group.Name.NamespaceName;
            var gs = group.Element(ns + "S_" + Segment.Gs);

            if (gs == null)
            {
                throw new ParserException("Can't find GS segment.");
            }

            this.ediSegments.AddRange(ParserHelper.ParseSegment(gs, ParserHelper.GetParsingGrammarForEnvelopeSegment<S_GS>(interchangeContext), interchangeContext));
        }

        public override void CreateSegment(XElement segment, XElement typedMessage, string origin)
        {
            var messageContext = CreateMessageContext(typedMessage, origin);
            var segmentGrammar =
                ParserHelper.GetParsingGrammarForSegment(
                    new ParseTree
                        {
                            TypeName = messageContext.TypeName + "." + segment.Name.LocalName,
                            AssemblyName = messageContext.AssemblyName
                        });

            this.ediSegments.AddRange(
                ParserHelper.ParseSegment(segment, segmentGrammar, interchangeContext));
        }

        public override void CreateGroupTrailer(XElement group)
        {
            XNamespace ns = group.Name.NamespaceName;
            var ge = group.Element(ns + "S_" + Segment.Ge);

            if (ge == null)
            {
                throw new ParserException("Can't find GE segment.");
            }

            this.ediSegments.AddRange(ParserHelper.ParseSegment(ge, ParserHelper.GetParsingGrammarForEnvelopeSegment<S_GE>(interchangeContext), interchangeContext));
        }

        public override void CreateInterchangeTrailer(XElement xmlEdi)
        {
            XNamespace ns = xmlEdi.Name.NamespaceName;
            var iea = xmlEdi.Element(ns + "S_" + Segment.Iea);

            if (iea == null)
            {
                throw new ParserException("Can't find IEA segment.");
            }

            this.ediSegments.AddRange(ParserHelper.ParseSegment(iea, ParserHelper.GetParsingGrammarForEnvelopeSegment<S_IEA>(interchangeContext), interchangeContext));
        }

        private MessageContext CreateMessageContext(XElement message, string origin)
        {
            var result = new MessageContext();

            XNamespace ns = message.Name.NamespaceName;
            var st = message.Element(ns + "S_" + Segment.St);

            if (st == null)
            {
                throw new ParserException("Can't find ST segment.");
            }
            
            if(message.Parent == null || message.Parent.Parent == null || message.Parent.Parent.Parent == null)
            {
                throw new ParserException("Can't find GROUP segment.");
            }
            
            var gs = message.Parent.Parent.Parent.Element(ns + "S_" + Segment.Gs);

            if (gs == null)
            {
                throw new ParserException("Can't find GS segment.");
            }

            var tag = st.Element(ns + Internals.StTag);

            if (tag == null)
            {
                throw new ParserException("Can't find " + Internals.StTag);
            }

            var version = gs.Element(ns + Internals.GsVersion);

            if (version == null)
            {
                throw new ParserException("Can't find " + Internals.GsVersion);
            }

            if (version.Value.Length < 6)
            {
                throw new ParserException("Version length should be at least 6.");
            }

            result.Tag = tag.Value;
            result.Version = version.Value.Substring(0, 6);
            result.Origin = origin;
            result.Format = Format.X12;

            return result;
        }
    }
}
