﻿//---------------------------------------------------------------------
// This file is part of ediFabric
//
// Copyright (c) ediFabric. All rights reserved.
//
// THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
// KIND, WHETHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
// PURPOSE.
//---------------------------------------------------------------------

namespace EdiFabric.Framework.Parsers.Interchange
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Xml.Linq;

    using EdiFabric.Framework.Constants;
    using EdiFabric.Framework.Entities;
    using EdiFabric.Framework.Exceptions;
    using EdiFabric.Framework.Helpers;
    using EdiFabric.Framework.InterchangeEnvelopes.X12;
    using EdiFabric.Framework.Parsers.Message;

    class InterchangeX12Builder : InterchangeEdiBuilder
    {
        private INTERCHANGE X12Envelope
        {
            get
            {
                return this.interchangeEnvelope as INTERCHANGE;
            }
        }

        public InterchangeX12Builder(InterchangeContext interchangeContext)
        {
            if (interchangeContext == null) throw new ArgumentNullException("interchangeContext");

            this.interchangeContext = interchangeContext;
            this.interchangeEnvelope = new INTERCHANGE { GROUPS = new List<GROUP>() };
        }

        public override void CreateInterchangeHeader(string segment)
        {
            X12Envelope.S_ISA = ParserHelper.PopulateSegment<S_ISA>(segment, this.interchangeContext);
        }

        public override void CreateInterchangeTrailer(string segment)
        {
            X12Envelope.S_IEA = ParserHelper.PopulateSegment<S_IEA>(segment, this.interchangeContext);
        }

        public override void CreateGroupHeader(string segment)
        {
            this.X12Envelope.GROUPS.Add(new GROUP());
            this.X12Envelope.GROUPS.Last().MESSAGES = new List<XElement>();
            this.X12Envelope.GROUPS.Last().S_GS = ParserHelper.PopulateSegment<S_GS>(segment, this.interchangeContext);
        }

        public override void CreateGroupTrailer(string segment)
        {
            this.X12Envelope.GROUPS.Last().S_GE = ParserHelper.PopulateSegment<S_GE>(segment, this.interchangeContext);
        }

        public override void CreateMessage(List<string> segments, MessageContext messageContext, string origin)
        {
            if (segments == null) throw new ArgumentNullException("segments");

            var parseTree = ParserHelper.GetParsingGrammarForMessage(messageContext);
            var director = new MessageDirector(segments, this.interchangeContext);
            var xmlEdi = director.Construct(parseTree);

            this.X12Envelope.GROUPS.Last().MESSAGES.Add(xmlEdi);
        }

        public override MessageContext GetMessageContext(IEnumerable<string> ediSegments, string origin)
        {
            if (ediSegments == null) throw new ArgumentNullException("ediSegments");
            
            var result = new MessageContext();
            var enumerable = ediSegments as List<string> ?? ediSegments.ToList();

            try
            {
                var gs = enumerable.Single(es => es.StartsWith(Segment.Gs));
                var ediCompositeDataElementsGs = EdiHelper.GetEdiCompositeDataElements(gs, this.interchangeContext);

                result.Version = ediCompositeDataElementsGs[7].Substring(0, 6);
            }
            catch
            {
                throw new ParserException("Can't parse GS segment.");
            }

            try
            {
                var st = enumerable.Single(es => es.StartsWith(Segment.St));
                var ediCompositeDataElementsSt = EdiHelper.GetEdiCompositeDataElements(st, this.interchangeContext);
                result.Tag = ediCompositeDataElementsSt[0];
                result.Origin = origin;
            }
            catch
            {
                throw new ParserException("Can't parse ST segment.");
            }

            result.Format = Format.X12;

            return result;
        }
    }
}
