﻿using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace EdiFabric.Framework.Tests
{
    using System;
    using System.Collections.Generic;
    using System.Diagnostics;
    using System.IO;
    using System.Reflection;
    using System.Xml.Schema;
    using System.Xml.Linq;

    using EdiFabric.Framework.InterchangeEnvelopes.X12;
    using EdiFabric.Framework.Parsers;
    using EdiFabric.Framework.Exceptions;

    [TestClass]
    public class TestX12
    {
        [TestMethod]
        public void TestToInterchangeWithEnvelopeSchemaValidation()
        {
            // ARRANGE
            const string Sample = "EdiFabric.Tests.Edi.X12_810_00204.txt";
            const string EnvelopeXsd = "EdiFabric.Tests.Xsd.X12Envelope.xsd";

            // ACT
            IEdiParser ediParser = new EdiParser();
            var interchange = ediParser.ToInterchange<INTERCHANGE>(Assembly.GetExecutingAssembly().GetManifestResourceStream(Sample), null);
            var xml = ediParser.Serialize(interchange);
            
            var schemas = new XmlSchemaSet();
            var stream = Assembly.GetExecutingAssembly().GetManifestResourceStream(EnvelopeXsd);
            if (stream != null)
            {
                var schema = XmlSchema.Read(stream, null);
                schemas.Add(schema);
            }
            var doc = new XDocument(xml);
            var errors = false;
            doc.Validate(schemas, (o, e) => { errors = true; });

            // ASSERT
            Assert.IsFalse(errors);
        }

        [TestMethod]
        public void TestToInterchangeWithNonDefaultSeparators()
        {
            // ARRANGE
            const string Sample = "EdiFabric.Tests.Edi.X12_810_00204_NonDefaultSeparators.txt";
            const string ExpectedResult = "EdiFabric.Tests.Xml.X12_810_00204.xml";

            var stream = Assembly.GetExecutingAssembly().GetManifestResourceStream(ExpectedResult);
            Debug.Assert(stream != null, "stream != null");
            var expectedXml = XElement.Load(stream);

            // ACT
            IEdiParser ediParser = new EdiParser();
            var interchange = ediParser.ToInterchange<INTERCHANGE>(Assembly.GetExecutingAssembly().GetManifestResourceStream(Sample), null);
            var parsedXml = ediParser.Serialize(interchange);

            // ASSERT
            Assert.AreEqual(parsedXml.ToString(), expectedXml.ToString());
        }

        [TestMethod]
        public void TestToInterchangeWithSegmentSeparatorIsNewLine()
        {
            // ARRANGE
            const string Sample = "EdiFabric.Tests.Edi.X12_810_00204_SegmentSeparatorIsNewLine.txt";
            const string ExpectedResult = "EdiFabric.Tests.Xml.X12_810_00204.xml";

            var stream = Assembly.GetExecutingAssembly().GetManifestResourceStream(ExpectedResult);
            Debug.Assert(stream != null, "stream != null");
            var expectedXml = XElement.Load(stream);

            // ACT
            IEdiParser ediParser = new EdiParser();
            var interchange = ediParser.ToInterchange<INTERCHANGE>(Assembly.GetExecutingAssembly().GetManifestResourceStream(Sample), null);
            var parsedXml = ediParser.Serialize(interchange);

            // ASSERT
            Assert.AreEqual(parsedXml.ToString(), expectedXml.ToString());
        }

        [TestMethod]
        public void TestToInterchangeWithXmlComparison()
        {
            // ARRANGE
            const string Sample = "EdiFabric.Tests.Edi.X12_810_00204.txt";
            const string ExpectedResult = "EdiFabric.Tests.Xml.X12_810_00204.xml";

            var stream = Assembly.GetExecutingAssembly().GetManifestResourceStream(ExpectedResult);
            Debug.Assert(stream != null, "stream != null");
            var expectedXml = XElement.Load(stream);

            // ACT
            IEdiParser ediParser = new EdiParser();
            var interchange = ediParser.ToInterchange<INTERCHANGE>(Assembly.GetExecutingAssembly().GetManifestResourceStream(Sample), null);
            var parsedXml = ediParser.Serialize(interchange);

            // ASSERT
            Assert.AreEqual(parsedXml.ToString(), expectedXml.ToString());
        }

        [TestMethod]
        public void TestToEdiWithSegmentsComparison()
        {
            // ARRANGE
            const string Sample = "EdiFabric.Tests.Xml.X12_810_00204.xml";
            const string ExpectedResult = "EdiFabric.Tests.Edi.X12_810_00204.txt";

            var stream = Assembly.GetExecutingAssembly().GetManifestResourceStream(ExpectedResult);
            Debug.Assert(stream != null, "stream != null");

            var reader = new StreamReader(stream);
            var expectedEdi = new List<string>();
            while (reader.Peek() >= 0)
            {
                expectedEdi.Add(reader.ReadLine());
            }

            // ACT
            IEdiParser ediParser = new EdiParser();
            var parsedEdi = ediParser.ToEdi(XElement.Load(Assembly.GetExecutingAssembly().GetManifestResourceStream(Sample)), null, null);

            // ASSERT
            Assert.AreEqual(expectedEdi.Count, parsedEdi.Count);
            for (int i = 0; i < parsedEdi.Count; i++)
            {
                Assert.IsTrue(parsedEdi[i] == expectedEdi[i]);
            }
        }

        [TestMethod]
        public void TestToEdiWithSegmentsComparisonAndNonDefaultSeparators()
        {
            // ARRANGE
            const string Sample = "EdiFabric.Tests.Xml.X12_810_00204.xml";
            const string ExpectedResult = "EdiFabric.Tests.Edi.X12_810_00204_NonDefaultSeparators.txt";

            var stream = Assembly.GetExecutingAssembly().GetManifestResourceStream(ExpectedResult);
            Debug.Assert(stream != null, "stream != null");

            var reader = new StreamReader(stream);
            var expectedEdi = new List<string>();
            while (reader.Peek() >= 0)
            {
                expectedEdi.Add(reader.ReadLine());
            }

            // ACT
            IEdiParser ediParser = new EdiParser();
            var interchangeContext = new Entities.InterchangeContext();
            interchangeContext.DataElementSeparator = "|";
            var parsedEdi = ediParser.ToEdi(XElement.Load(Assembly.GetExecutingAssembly().GetManifestResourceStream(Sample)), null, interchangeContext);

            // ASSERT
            Assert.AreEqual(expectedEdi.Count, parsedEdi.Count);
            for (int i = 0; i < parsedEdi.Count; i++)
            {
                Assert.IsTrue(parsedEdi[i] == expectedEdi[i]);
            }
        }

        [TestMethod]
        public void TestToInterchangeWithFriendlyException()
        {
            // ARRANGE
            const string Sample = "EdiFabric.Tests.Edi.X12_810_00204_BadSegment.txt";
           
            // ACT
            try
            {
                IEdiParser ediParser = new EdiParser();
                var interchange = ediParser.ToInterchange<INTERCHANGE>(Assembly.GetExecutingAssembly().GetManifestResourceStream(Sample), null);
            }
            catch (Exception ex)
            {
                // ASSERT
                Assert.IsInstanceOfType(ex, typeof(ParserException));
            }
        }

        [TestMethod]
        public void TestToInterchangeWithMultipleMessages()
        {
            // ARRANGE
            const string Sample = "EdiFabric.Tests.Edi.X12_810_00204_MultipleMessages.txt";
            const string ExpectedResult = "EdiFabric.Tests.Xml.X12_810_00204_MultipleMessages.xml";

            var stream = Assembly.GetExecutingAssembly().GetManifestResourceStream(ExpectedResult);
            Debug.Assert(stream != null, "stream != null");
            var expectedXml = XElement.Load(stream);

            // ACT
            IEdiParser ediParser = new EdiParser();
            var interchange = ediParser.ToInterchange<INTERCHANGE>(Assembly.GetExecutingAssembly().GetManifestResourceStream(Sample), null);
            var parsedXml = ediParser.Serialize(interchange);

            // ASSERT
            Assert.AreEqual(parsedXml.ToString(), expectedXml.ToString());
        }

        [TestMethod]
        public void TestToInterchangeWithMultipleGroups()
        {
            // ARRANGE
            const string Sample = "EdiFabric.Tests.Edi.X12_810_00204_MultipleGroups.txt";
            const string ExpectedResult = "EdiFabric.Tests.Xml.X12_810_00204_MultipleGroups.xml";

            var stream = Assembly.GetExecutingAssembly().GetManifestResourceStream(ExpectedResult);
            Debug.Assert(stream != null, "stream != null");
            var expectedXml = XElement.Load(stream);

            // ACT
            IEdiParser ediParser = new EdiParser();
            var interchange = ediParser.ToInterchange<INTERCHANGE>(Assembly.GetExecutingAssembly().GetManifestResourceStream(Sample), null);
            var parsedXml = ediParser.Serialize(interchange);

            // ASSERT
            Assert.AreEqual(parsedXml.ToString(), expectedXml.ToString());
        }

        [TestMethod]
        public void TestToInterchangeWithOrigin()
        {
            // ARRANGE
            const string Sample = "EdiFabric.Tests.Edi.X12_810_00204_Origin.txt";
            const string ExpectedResult = "EdiFabric.Tests.Xml.X12_810_00204_Origin.xml";

            var stream = Assembly.GetExecutingAssembly().GetManifestResourceStream(ExpectedResult);
            Debug.Assert(stream != null, "stream != null");
            var expectedXml = XElement.Load(stream);

            // ACT
            IEdiParser ediParser = new EdiParser();
            var origin = "Custom";
            var interchange = ediParser.ToInterchange<INTERCHANGE>(Assembly.GetExecutingAssembly().GetManifestResourceStream(Sample), origin);
            var parsedXml = ediParser.Serialize(interchange);

            // ASSERT
            Assert.AreEqual(parsedXml.ToString(), expectedXml.ToString());
        }
    }
}
